from .closedrows import ClosedRows, PersistentClosedRows
from .index import Index, PersistentIndex
from .timeconfig import TimeConfig, PersistentTimeConfig
from .warmingupcount import WarmingUpCount, PersistentWarmingUpCount


__all__ = [
    "ClosedRows",
    "PersistentClosedRows",
    "Index",
    "PersistentIndex",
    "TimeConfig",
    "PersistentTimeConfig",
    "WarmingUpCount",
    "PersistentWarmingUpCount",
]



class DuplicateError(RuntimeError):
    """
    Raised when the key already exists in the same document.
    """

    def __init__(self, message, obj_id, key_value=None):
        super().__init__(message)
        self.ObjId = obj_id
        self.KeyValue = key_value



from bspump.asab import Config

from ..abc.anomaly import Anomaly


class GeneralAnomaly(Anomaly):
    """
    GeneralAnomaly serves a default anomaly class.

    Unless a specific AnomalyObject is created for a given type of the anomaly/symptom,
    GeneralAnomaly is going to be used as default.

    close_rules define how each anomaly type is closed, i. e. based on "status", or time (specify the value in seconds)
    """

    TYPE = "default"  # N/A

    Config.add_defaults(
        {
            "GeneralAnomaly": {
                "close_rules": "default:status",  # Structure: type1:rule;type2:rule ...
            },
        }
    )

    def __init__(self):
        super().__init__()

        self.CloseRules = {}
        for key_value_str in str(Config["GeneralAnomaly"]["close_rules"]).split(";"):
            key_value = key_value_str.split(":")
            self.CloseRules[key_value[0]] = key_value[1]

    async def on_tick(self, current_time):
        if self["status"] == "closed":
            return

        close_rule = self.CloseRules.get(self["type"], self.CloseRules.get("default"))

        # Obtain last symptom
        last_timestamp = 0
        last_status = "open"
        for symptom in self["symptoms"]:
            timestamp = symptom.get("@timestamp")
            if timestamp > last_timestamp:
                last_timestamp = timestamp
                last_status = symptom.get("status", "open")

        if close_rule == "status" and last_status == "closed":
            self["status"] = "closed"

        if close_rule != "status":
            close_rule_seconds = int(close_rule)
            if last_timestamp != 0 and current_time > (
                last_timestamp + close_rule_seconds
            ):
                self["status"] = "closed"



from ..abc.lookup import Lookup
import pymemcache.client
import logging

L = logging.getLogger(__name__)

###


class MemcachedLookup(Lookup):
    ConfigDefaults = {
        "max_size": 1000,
        "expiry_seconds": 3600,
        "server": "127.0.0.1:11211",
    }

    def __init__(self, app, id=None, config=None):
        super().__init__(app, id=id, config=config)

        self.App = app
        self.MaxSize = self.Config["max_size"]
        self.Expiration = self.Config["expiry_seconds"]
        self.Server = self.Config["server"]
        self.Target = None

        if len(self.Server) > 0:
            self.Client = pymemcache.client.base.Client(tuple(self.Server.split(":")))
        else:
            raise Exception("Memcache service path not set.")

    def set(self, to_chache: dict):
        for key, value in to_chache.items():
            returned = self.Client.set(key, value, expire=self.Expiration)
        if returned is not True:
            L.warning("Setting a memcached key-value failed")
        L.debug("Successfully set key-value")

    def get(self, key):
        return self.Client.get(key)

    def delete(self, key):
        self.Client.delete(key)
        L.debug("Deleted {} form memcache.".format(key))

    def close_connection(self):
        self.Client.close()
        L.info("Closed connection to memcache")

    def flush_all(self):
        self.Client.flush_all()
        L.info("Chache flushed")



#!/usr/bin/env python3
import logging
import io
import fastavro
from ..avro import loader
from ..abc.generator import Generator


###

L = logging.getLogger(__name__)

###


class AvroDeserializer(Generator):
    ConfigDefaults = {
        "schema_file": "",
    }

    def __init__(self, app, pipeline, id=None, config=None):
        super().__init__(app, pipeline, id=id, config=config)
        self.Schema = loader.load_avro_schema(self.Config)

    async def generate(self, context, event, depth):
        fi = io.BytesIO(event)

        if self.Schema is None:
            L.warning("Schema file is not provided.")
        else:
            L.warning("Schema file is used.")

        for record in fastavro.reader(fi, self.Schema):
            self.Pipeline.inject(context, record, depth)



import json
from ..abc.processor import Processor


class StdDictToJsonParser(Processor):
    """
    Description:

    |

    """

    def process(self, context, event):
        """
        Description:

        :return: ?

        |

        """
        return json.dumps(event)


class StdJsonToDictParser(Processor):
    """
    Description:

    |

    """

    def process(self, context, event):
        """
        Description:

        :return: ???

        |

        """
        return json.loads(event)


class DictToJsonBytesParser(Processor):
    """
    DictToJsonBytesParser transforms a dictionary to JSON-string encoded in bytes.
    The encoding charset can be specified in the configuration in `encoding` field.

    |

    """

    ConfigDefaults = {
        "encoding": "utf-8",
    }

    def __init__(self, app, pipeline, id=None, config=None):
        """
        Description: ..

        |

        """
        super().__init__(app, pipeline, id, config)
        self.Encoding = self.Config["encoding"]

    def process(self, context, event):
        """
        Description:

        :return: ??

        |

        """
        assert isinstance(event, dict)
        return json.dumps(event).encode(self.Encoding)



# Watch the files for changes and restart if they changed
import os
import sys
from pathlib import Path

from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler


def get_imported_files():
    imported_files = set()

    for name, module in sys.modules.items():
        try:
            # Skip modules without __file__ attribute (like built-in modules)
            if hasattr(module, "__file__") and module.__file__:
                # Convert to absolute path and resolve any symlinks
                file_path = Path(module.__file__).resolve()

                # Only include .py files (skip .pyc, .pyd, .so files)
                if file_path.suffix == ".py":
                    if "site-packages" not in str(
                        file_path
                    ) and "lib/python" not in str(file_path):
                        imported_files.add(str(file_path))
        except (AttributeError, TypeError):
            continue

    return list(imported_files)


class Watcher:
    def __init__(self, files):
        self.files = files + get_imported_files()
        self.observer = Observer()
        self.event_handler = WatcherEventHandler(self)
        for file in self.files:
            self.observer.schedule(self.event_handler, file, recursive=False)
        self.observer.start()


class WatcherEventHandler(FileSystemEventHandler):
    def __init__(self, watcher):
        self.watcher = watcher

    def on_modified(self, event):
        if event.is_directory:
            return
        if event.src_path in self.watcher.files:
            # exec yourself with the same args
            os.execv(sys.argv[0], sys.argv)



from .sink import UnitTestSink
from .source import UnitTestSource
from ..pipeline import Pipeline
from ..trigger import PubSubTrigger
from ..abc.processor import Processor
from ..application import BSPumpApplication


class UnitTestPipeline(Pipeline):
    def __init__(
        self, app: type(BSPumpApplication), processor: type(Processor), *args, **kwargs
    ):
        """
        Build Pipeline with given app and processor

        :param processor: Processor to test - available as self.Processor
        :param args: Optional arguments for processor
        :param kwargs: Optional key-word arguments for processor
        """
        super().__init__(app, "UnitTestPipeline")

        self.Source = UnitTestSource(app, self).on(
            PubSubTrigger(app, "Application.run!", app.PubSub)
        )
        self.Processor = processor(app, self, *args, **kwargs)
        self.Sink = UnitTestSink(app, self)

        self.PubSub.subscribe("bspump.pipeline.cycle_end!", self._on_finished)

        self.build(self.Source, self.Processor, self.Sink)

    def unittest_start(self):
        """
        Start the pipeline
        """
        pass

    def _on_finished(self, event_name, pipeline):
        self.App.stop()



import asyncio
import logging

from ..abc.source import Source

L = logging.getLogger(__name__)


class MQTTSource(Source):
    ConfigDefaults = {"topic": "#", "qos": 0}

    def __init__(self, app, pipeline, connection, id=None, config=None):
        super().__init__(app, pipeline, id=id, config=config)

        self.Connection = pipeline.locate_connection(app, connection)
        self.Connection.subscribe_topic(self.Config["topic"], self.Config["qos"])
        self.Connection.register_handler(self.Config["topic"], self.on_message)

        self._queue = asyncio.Queue()

    async def main(self):
        try:
            while True:
                await self.Pipeline.ready()
                event = await self._queue.get()
                await self.process(event)
        except asyncio.CancelledError:
            pass

        except BaseException as e:
            L.exception("Error when processing message.")
            self.Pipeline.set_error(None, None, e)

    def on_message(self, client, userdata, message):
        self._queue.put_nowait(message.payload)



import json
import yaml
import importlib
import glob
import os
import logging

import bspump.asab as asab

import bspump.analyzer

##

L = logging.getLogger(__name__)

##

asab.Config.add_defaults(
    {
        "SegmentBuilder": {
            "path": "./etc/processors/*",
        },
    }
)


class SegmentBuilder(object):
    """
    SegmentBuilder is meant to add a new segment to the pipeline,
    that consists of processors such as enrichers, analyzers etc.

    Usage:

            self.SegmentBuilder = bspump.declarative.SegmentBuilder()
            self.SegmentBuilder.construct_segment(self)

    The configuration file:

            pipeline_id: AnalyticsIPPipeline
            processor: |
                    --- !DICT
                    with: !EVENT
                    add:
                            local_ip:
                                    !IF
                                    is:
                                            !INSUBNET
                                            subnet: 192.168.0.0/16
                                            value:
                                                    !ITEM
                                                    with: !EVENT
                                                    item: CEF.sourceAddress
                                    then: true
                                    else: false
    """

    PROCESSORS = {
        "processor": ("bspump.declarative", "DeclarativeProcessor"),
        "generator": ("bspump.declarative", "DeclarativeGenerator"),
        "lookup": ("bspump.mongodb", "MongoDBLookup"),
    }

    def __init__(self):
        self.Path = asab.Config["SegmentBuilder"]["path"]
        self.DefinitionsLookups = []
        self.DefinitionsProcessors = []
        file_list = glob.glob(self.Path, recursive=True)
        file_list.sort()
        while len(file_list) > 0:
            file_name = file_list.pop(0)
            if os.path.isfile(file_name):
                with open(file_name) as file:
                    if file_name.endswith(".yml"):
                        # Read YML file
                        definition = yaml.load(file)
                    else:
                        # Read JSON file
                        definition = json.load(file)
                    if definition.get("lookup") is not None or definition.get(
                        "class", ""
                    ).endswith("Lookup"):
                        self.DefinitionsLookups.append(definition)
                    else:
                        self.DefinitionsProcessors.append(definition)

    def register_processor(self, processor_keyword, _module, _class):
        """
        Registers custom processors that are translated to module & class.
        The default keywords may also be overriden.
        """
        self.PROCESSORS[processor_keyword] = (_module, _class)

    def _map_processor(self, definition):
        for processor_keyword in self.PROCESSORS:
            if processor_keyword in definition:
                _module, _class = self.PROCESSORS[processor_keyword]
                definition["declaration"] = definition[processor_keyword]
                definition["module"] = _module
                definition["class"] = _class
        return definition

    def construct_segment(self, app):
        svc = app.get_service("bspump.PumpService")

        # First create lookups
        for definition in self.DefinitionsLookups:
            lookup = self.construct_lookup(app, self._map_processor(definition))
            svc.add_lookup(lookup)

        # Then create other processors
        for definition in self.DefinitionsProcessors:
            pipeline = svc.locate(definition["pipeline_id"])
            processor = self.construct_processor(
                app, pipeline, self._map_processor(definition)
            )
            if processor is not None:
                insert_before = definition.get("insert_before")
                if insert_before is not None:
                    pipeline.insert_before(id=insert_before, processor=processor)
                else:
                    sink = pipeline.Processors[-1].pop()
                    pipeline.append_processor(processor)
                    pipeline.append_processor(sink)
                self.build_profiling(pipeline, processor)

    def construct_lookup(self, app, definition):
        _module = importlib.import_module(definition["module"])
        _class = getattr(_module, definition["class"])
        lookup = _class.construct(app, definition)
        return lookup

    def construct_processor(self, app, pipeline, definition):
        _module = importlib.import_module(definition["module"])
        _class = getattr(_module, definition["class"])
        processor = _class.construct(app, pipeline, definition)
        return processor

    def build_profiling(self, pipeline, processor):
        pipeline.ProfilerCounter[processor.Id] = pipeline.MetricsService.create_counter(
            "bspump.pipeline.profiler",
            tags={
                "processor": processor.Id,
                "pipeline": pipeline.Id,
            },
            init_values={"duration": 0.0, "run": 0},
            reset=pipeline.ResetProfiler,
        )
        if isinstance(processor, bspump.analyzer.Analyzer):
            pipeline.ProfilerCounter[
                "analyzer_" + processor.Id
            ] = pipeline.MetricsService.create_counter(
                "bspump.pipeline.profiler",
                tags={
                    "analyzer": processor.Id,
                    "pipeline": pipeline.Id,
                },
                init_values={"duration": 0.0, "run": 0},
                reset=pipeline.ResetProfiler,
            )



import base64
from io import BytesIO

from aiohttp.web import Request

from bspump.http.web.components.field import Field
from bspump.http.web.template_env import env


class FileField(Field):
    """
    The value ends up being the bytes of the uploaded file.
    """

    def inner_html(self, default="", readonly=False):
        template = env.get_template("file-field.html")
        return template.render(
            default_input_props=self.default_input_props,
            default_classes=self.default_classes,
        )

    def clean(self, data, request: Request | None = None):
        if request.content_type == "application/json":
            decoded_data = base64.b64decode(data.get(self.name, ""))
            data[self.name] = BytesIO(decoded_data)
        else:
            # in case of not submitting any file
            if data[self.name] == b"":
                data[self.name] = BytesIO(b"")
            else:
                data[self.name] = data[self.name].file



import re
import logging

import aiohttp

from ...abc.service import Service
from ...config import Config

#

L = logging.getLogger(__name__)

#


# "tenant_url" is used to periodically refresh tenants from, expecting "_id" inside a JSON structure,
# which is compatible with SeaCat Auth product
Config.add_defaults(
    {
        "tenants": {
            "ids": "",  # List of tenant ids, entries can be separated by comma or newline
            "tenant_url": "",  # f. e. http://seacat-auth:8080/tenant
            "trusted": 0,  # makes sure the tenants are implicitly trusted, even though they are not located in IDs or tenant URL
        }
    }
)


class TenantService(Service):
    def __init__(self, app, service_name="asab.TenantService"):
        super().__init__(app, service_name)
        self.App = app
        self.TenantWebHandler = None
        self.TenantsTrusted = Config.getboolean("tenants", "trusted")
        self.Tenants = set()

        # Load tenants from configuration
        for tenant_id in re.split(
            r"[,\s]+", Config.get("tenants", "ids"), flags=re.MULTILINE
        ):
            tenant_id = tenant_id.strip()
            # Skip comments and empty lines
            if len(tenant_id) == 0:
                continue
            if tenant_id[0] == "#":
                continue
            if tenant_id[0] == ";":
                continue
            self.Tenants.add(tenant_id)

        # Load tenants from URL
        self.TenantUrl = Config.get("tenants", "tenant_url")

    async def initialize(self, app):
        if len(self.TenantUrl) > 0:
            await self._update_tenants()
            # TODO: Websocket persistent API should be added to seacat auth to feed these changes in realtime (eventually)
            app.PubSub.subscribe("Application.tick/300!", self._update_tenants)

    def locate_tenant(self, tenant_id):
        if tenant_id in self.Tenants:
            return tenant_id
        elif self.TenantsTrusted:
            self.Tenants.add(tenant_id)
            return tenant_id
        else:
            return None

    def get_tenant_ids(self):
        # TODO: REMOVE?
        return self.get_tenants()

    def get_tenants(self):
        return list(self.Tenants)

    def add_web_api(self, web_container):
        from .web import TenantWebHandler

        self.TenantWebHandler = TenantWebHandler(self.App, self, web_container)

    async def _update_tenants(self, message_name=None):
        async with aiohttp.ClientSession() as session:
            async with session.get(self.TenantUrl) as resp:
                if resp.status == 200:
                    tenants_list = await resp.json()
                    for tenant in tenants_list:
                        if isinstance(tenant, str):
                            self.Tenants.add(tenant)
                        elif isinstance(tenant, dict) and "_id" in tenant:
                            # TODO: REMOVE?
                            self.Tenants.add(tenant["_id"])
                        else:
                            L.warning("Unknown tenant format: {}".format(tenant))



import logging
import asyncio
import asyncssh

from ..abc.connection import Connection


#

L = logging.getLogger(__name__)

#


class SSHConnection(Connection):
    """

            SSHConnection is used to connect SFTPSink to the SFTP server and upload files to the remote folder.

            SSHConnection is built on top of asyncssh library and utilizes its functions: https://asyncssh.readthedocs.io/en/latest/

            The following code illustrates how to create and register the SSH connection inside the application object.

    .. code:: python


            app = bspump.BSPumpApplication()
            svc = app.get_service("bspump.PumpService")

            svc.add_connection(
                    bspump.ssh.SSHConnection(app, "SSHConnection")
            )

    ..

            'ConfigDefaults' footnotes:

            'known_hosts'
            The list of keys which will be used to validate the server host key presented during the SSH handshake.If this is
            explicitly set to '' or None, server host key validation will be disabled.

            'client_host_keysign'
            Whether or not to use ssh-keysign to sign host-based authentication requests. If set to 'true', an attempt will be
            made to find ssh-keysign in its typical locations. If set to a string, that will be used as the ssh-keysign path.
            When set, client_host_keys should be a list of public keys. Otherwise, client_host_keys should be a list of private
            keys with optional paired certificates

            'client_host_keys'
            A list of keys to use to authenticate this client via host-based authentication. If client_host_keysign is set and no
            host keys or certificates are specified, an attempt will be made to find them in their typical locations. If
            client_host_keysign is not set, host private keys must be specified explicitly or host-based authentication will not be
            performed.

    """

    ConfigDefaults = {
        "host": "localhost",
        "port": 22,
        "user": "",
        "password": "",
        "known_hosts": "",  # separate paths / files with comma
        "client_host_keysign": "",  # '' -> False; 'true' -> True; 'anything_else' -> path to the directory with keys
        "client_host_keys": "",  # '' -> default. If set, separate keynames with comma
    }

    def __init__(self, app, id=None, config=None):
        super().__init__(app, id=id, config=config)

        self.ConnectionEvent = asyncio.Event()
        self.ConnectionEvent.clear()

        self.Loop = app.Loop

        self.Host = self.Config["host"]
        self.Port = int(self.Config["port"])
        self.User = self.Config["user"]
        self.Password = self.Config["password"]
        self.Known_hosts = list(self.Config["known_hosts"].split(","))
        self.Client_keysign = self.Config["client_host_keysign"]
        self.Client_keys = list(self.Config["client_host_keys"].split(","))

    def run(self):
        # Checks void known hosts and client host keys
        if not self.Known_hosts or self.Known_hosts == [""] or self.Known_hosts is None:
            self.Known_hosts = None
        if not self.Client_keys or self.Client_keys == [""] or self.Client_keys is None:
            self.Client_keys = None
        # Sets client host keysign to boolean value. If set to path, it will try to find keysign in given path
        if self.Client_keysign != "" and self.Client_keysign != "true":
            self.Client_keysign = str(self.Client_keysign)
        else:
            self.Client_keysign = bool(self.Client_keysign)

        conn = asyncssh.connect(
            host=self.Host,
            port=self.Port,
            username=self.User,
            password=self.Password,
            known_hosts=self.Known_hosts,
            client_host_keysign=self.Client_keysign,
            client_host_keys=self.Client_keys,
        )

        return conn



from ..abc.sink import Sink


class UnitTestSink(Sink):
    def __init__(self, app, pipeline, id=None, config=None):
        super().__init__(app, pipeline, id=id, config=config)
        self.Output = []

    def process(self, context, event):
        self.Output.append((context, event))



import logging
import io
import apiclient.http
from .abcsink import GoogleDriveABCSink


L = logging.getLogger(__name__)


class GoogleDriveBlockSink(GoogleDriveABCSink):
    """
    GoogleDriveBlockSink is a file sink processor that can be used to upload individual
    events as files to Google drive using GoogleDriveConnection.
    Properties `filename` and 'mimetype' from event context are used as metadata.
    If they are not present, default from configuration is used instead.
    """

    ConfigDefaults = {
        "parent_folder_id": "",
        "default_filename": "untitled_file",
        "default_mimetype": "",
    }

    def __init__(self, app, pipeline, connection, id=None, config=None):
        super().__init__(app, pipeline, connection, id=id, config=config)

        self.ParentFolderID = self.Config["parent_folder_id"]
        if self.ParentFolderID == "":
            self.ParentFolderID = None
        self.DefaultFilename = self.Config["default_filename"]
        self.DefaultMimeType = self.Config["default_mimetype"]

    async def _loader(self):
        while True:
            event, context = await self._output_queue.get()

            if event is None:
                break

            ctx_mime = context.get("mimetype")
            mimetype = ctx_mime if ctx_mime is not None else self.DefaultMimeType

            ctx_fn = context.get("filename")
            filename = ctx_fn if ctx_fn is not None else self.DefaultFilename

            if self._output_queue.qsize() == self.Connection._output_queue_max_size - 1:
                self.Pipeline.throttle(self, False)

            file_metadata = {"name": filename, "parents": [self.ParentFolderID]}
            file = io.BytesIO(event)
            media = apiclient.http.MediaIoBaseUpload(
                file, mimetype=mimetype, resumable=True
            )
            drive_file = (
                self.Drive_service.files()
                .create(body=file_metadata, media_body=media, fields="id")
                .execute()
            )

            if drive_file.get("id") is None:
                L.error(f"Failed upload file to Google drive: {file_metadata['name']}")



"""Asynchronous Server Application Boilerplate.

ASAB (Asynchronous Serverless Application Builder) is a lightweight Python framework for building
asynchronous and serverless applications. It provides a simple yet powerful set of tools and
conventions for developing scalable and efficient applications.

For detailed documentation and examples, please refer to the ASAB documentation at:
	- GitHub Repository: https://github.com/TeskaLabs/asab
	- Documentation: https://asab.readthedocs.io
"""

import logging

from .abc.module import Module
from .abc.service import Service
from .abc.singleton import Singleton
from .application import Application
from .config import Config, ConfigObject, Configurable
from .log import LOG_NOTICE
from .pdict import PersistentDict
from .pubsub import subscribe, PubSub, Subscriber
from .socket import StreamSocketServerService
from .timer import Timer

from .__version__ import __version__, __build__

# This logger is used to indicate use of the obsolete function
# Example:
#    asab.LogObsolete.warning("Use of the obsolete function", struct_data={'eol':'2022-31-12'})
# See https://github.com/TeskaLabs/asab/issues/240
LogObsolete = logging.getLogger("OBSOLETE")

__all__ = (
    "Module",
    "Service",
    "Singleton",
    "Application",
    "Config",
    "Configurable",
    "ConfigObject",
    "LOG_NOTICE",
    "PersistentDict",
    "subscribe",
    "PubSub",
    "Subscriber",
    "StreamSocketServerService",
    "Timer",
    "__version__",
    "__build__",
)



from ...abc import Expression
from ..value.valueexpr import VALUE


class CONTEXT(Expression):
    """
    The current context.

    Usage:
    ```
    !CONTEXT
    ``
    """

    Attributes = {}

    def __init__(self, app, *, value):
        super().__init__(app)
        assert value == ""

    def __call__(self, context, event, *args, **kwargs):
        return context


class CONTEXT_SET(Expression):
    Attributes = {
        "What": ["*"],  # TODO: This ...
    }

    def __init__(self, app, *, arg_set, arg_what=None):
        super().__init__(app)
        self.What = arg_what

        self.ArgSet = arg_set
        self.Set = None

    def set(self, key, value):
        setattr(self, key, value)

        if "Set" in key:
            self.Set[key[3:]] = value

    def initialize(self):
        if self.ArgSet is None:
            self.Set = None
        else:
            assert isinstance(self.ArgSet, dict)
            self.Set = dict()
            self._set_value_or_expression_to_attribute(self.ArgSet, self.Set, "Set")

    def _set_value_or_expression_to_attribute(self, _from, _to, name):
        for key, value in _from.items():
            # Parse value to attribute, if not expression
            if isinstance(value, Expression):
                _to[key] = value
            else:
                assert (
                    isinstance(value, (int, str, bytes, bool, tuple, list))
                    or value is None
                )
                value = VALUE(self.App, value=value)
                _to[key] = value

            # Set the value to the class attribute
            attr_name = "{}{}".format(name, key)
            setattr(self, attr_name, value)
            self.Attributes[attr_name] = value.get_outlet_type()

    def __call__(self, context, event, *args, **kwargs):
        if self.Set is not None:
            for key, value in self.Set.items():
                v = value(context, event, *args, **kwargs)
                if v is not None:
                    context[key] = v

        # TODO: Optimize
        if self.What is None:
            return None

        return self.What(context, event, *args, **kwargs)



import uuid
import logging

from .storage import SessionStorage
from .cookies import CookieSessionMixIn


L = logging.getLogger(__name__)


class InMemorySessionStorage(SessionStorage, CookieSessionMixIn):
    def __init__(self, app, session_class, max_age=60 * 15, cookie_name="SESSID"):
        super().__init__(app, max_age=max_age, session_class=session_class)
        self.set_cookie_name(cookie_name)

        self.Sessions = {}

        app.PubSub.subscribe("Application.tick/10!", self._on_tick)

    async def load(self, request):
        session_id = self.get_session_id_from_request(request)
        if session_id is not None:
            session = self.Sessions.get(session_id)
            if session is not None:
                if session.is_expired():
                    session = None
                else:
                    return session

            L.warning("Invalid session id '{}'".format(session_id))

        return await self.create(request)

    async def set(self, session, response):
        if session.is_new():
            while True:
                session_id = uuid.uuid4().hex
                if session_id not in self.Sessions:
                    session.set_id(session_id)
                    break

            self.Sessions[session.Id] = session
            self.set_session_to_response(session, response)

    async def delete(self, session):
        session_p = self.Sessions.pop(session.Id)
        assert session_p == session

    async def store(self, session):
        session.reset()

    async def _on_tick(self, message_type):
        # Find expired session and remove them
        expired_sessions = []
        for session in self.Sessions.values():
            if session.is_expired():
                expired_sessions.append(session)

        for session in expired_sessions:
            await self.delete(session)



import asyncio
import logging

import pika
import pika.adapters.asyncio_connection

from asab import PubSub

from ..abc.connection import Connection


L = logging.getLogger(__name__)


class AMQPConnection(Connection):
    ConfigDefaults = {
        "url": "amqp://localhost/",
        "appname": "bspump.py",
        "reconnect_delay": 10.0,
    }

    def __init__(self, app, id=None, config=None):
        super().__init__(app, id=id, config=config)

        self.Connection = None
        self.ConnectionEvent = asyncio.Event()
        self.ConnectionEvent.clear()

        self.PubSub = PubSub(app)
        self.Loop = app.Loop

        self._reconnect()

    def _reconnect(self):
        if self.Connection is not None:
            if not (self.Connection.is_closing or self.Connection.is_closed):
                self.Connection.close()
            self.Connection = None

        parameters = pika.URLParameters(self.Config["url"])
        if parameters.client_properties is None:
            parameters.client_properties = dict()
        parameters.client_properties["application"] = self.Config["appname"]

        self.Connection = pika.adapters.asyncio_connection.AsyncioConnection(
            parameters=parameters,
            on_open_callback=self._on_connection_open,
            on_open_error_callback=self._on_connection_open_error,
            on_close_callback=self._on_connection_close,
        )

    # Connection callbacks

    def _on_connection_open(self, connection):
        L.info("AMQP connected")
        self.ConnectionEvent.set()
        self.PubSub.publish("AMQPConnection.open!")

    def _on_connection_close(self, connection, code, reason=None):
        L.warning("AMQP disconnected ({}): {}".format(code, reason))
        self.ConnectionEvent.clear()
        self.PubSub.publish("AMQPConnection.close!")
        self.Loop.call_later(float(self.Config["reconnect_delay"]), self._reconnect)

    def _on_connection_open_error(self, connection, error_message=None):
        L.error(
            "AMQP error: {}".format(
                error_message if error_message is not None else "Generic error"
            )
        )
        self.ConnectionEvent.clear()
        self.PubSub.publish("AMQPConnection.close!")
        self.Loop.call_later(float(self.Config["reconnect_delay"]), self._reconnect)



import logging

import time

import bspump.asab as asab
from ..abc.processor import Processor

###

L = logging.getLogger(__name__)

###


class Analyzer(Processor):
    """
    Description:

    """

    ConfigDefaults = {
        "analyze_period": 60,  # every 60 seconds
    }

    def __init__(self, app, pipeline, analyze_on_clock=False, id=None, config=None):
        super().__init__(app, pipeline, id=id, config=config)
        self.AnalyzePeriod = float(self.Config["analyze_period"])
        self.AnalyzeOnClock = analyze_on_clock

        if analyze_on_clock:
            self.Timer = asab.Timer(app, self.on_clock_tick, autorestart=True)
            app.PubSub.subscribe("Application.run!", self.start_timer)
        else:
            self.Timer = None

    # Implementation interface

    def start_timer(self, event_type):
        """
        Description:

        """
        self.Timer.start(self.AnalyzePeriod)

    def analyze(self):
        """
        Description:

        """
        pass

    def evaluate(self, context, event):
        """
        The function which records the information from the event into the analyzed object.
                        Specific for each analyzer.

        **Parameters**

        context :

        event : any data type
                        information with timestamp.
        """
        pass

    def predicate(self, context, event):
        """
        This function is meant to check, if the event is worth to process.
        If it is, should return True.
        specific for each analyzer, but default one always returns True.

        **Parameters**

        context :

        event : any data type
                        information with timestamp.

        :return: True
        """
        return True

    def process(self, context, event):
        """
        The event passes through `process(context, event)` unchanged.
                        Meanwhile it is evaluated.

        **Parameters**

        context :

        event : any data type
                        information with timestamp.

        :return: event
        """
        if self.predicate(context, event):
            self.evaluate(context, event)

        return event

    async def on_clock_tick(self):
        """
        Run analyzis every tick.

        """
        t0 = time.perf_counter()
        self.analyze()
        self.Pipeline.ProfilerCounter["analyzer_" + self.Id].add(
            "duration", time.perf_counter() - t0
        )
        self.Pipeline.ProfilerCounter["analyzer_" + self.Id].add("run", 1)



import abc
import json
import urllib.parse
import uuid
import hashlib
import datetime
import logging
from bspump.asab.web.rest.json import JSONDumper
import http.client
import typing

#

L = logging.getLogger(__name__)

#


class UpsertorABC(abc.ABC):
    def __init__(self, storage, collection, obj_id, version=None):
        self.Storage = storage
        self.Collection = collection
        self.ObjId = obj_id

        self.Version = version

        now = datetime.datetime.now(datetime.timezone.utc)
        self.ModSet = {
            "_m": now,  # Set the modification datetime
        }
        if version == 0:
            self.ModSet["_c"] = now  # Set the creation datetime

        self.ModUnset = {}

        self.ModInc = {
            "_v": 1,  # Increment '_v' at every change
        }

        self.ModPush = {}
        self.ModPull = {}

        self.WebhookResponseData = {}

    def get_id_name(self):
        return "_id"

    @classmethod
    def generate_id(cls) -> bytes:
        """
        Generate a unique ID string using a combination of a random UUID and a SHA-256 hash.

        Returns:
                A string representation of the generated ID.
        """
        m = hashlib.sha256()
        m.update(uuid.uuid4().bytes)
        return m.digest()

    def set(self, objField, value, encrypt=False, encrypt_iv=None):
        """
        Add key and value to the upsertor.

        Args:
                objField: Key of the object.
                value: Value of the object.
                encrypt: Allow encryption.
                encrypt_iv: Custom initialization vector.
        """

        if encrypt:
            value = self.Storage.aes_encrypt(value, iv=encrypt_iv)
        self.ModSet[objField] = value

    def unset(self, obj_field):
        """
        Scalar unset
        """
        self.ModUnset[obj_field] = ""

    def increment(self, field_name, amount=1):
        """
        Scalar increment
        """
        self.ModInc[field_name] = amount

    def decrement(self, field_name, amount=1):
        """
        Scalar decrement
        """
        return self.increment(field_name, -amount)

    def push(self, field_name, value):
        """
        Push an item into a list
        """
        if self.ModPush.get(field_name) is None:
            self.ModPush[field_name] = []
        self.ModPush[field_name].append(value)

    def pull(self, field_name, value):
        """
        Pull an item from a list
        """
        if self.ModPull.get(field_name) is None:
            self.ModPull[field_name] = []
        self.ModPull[field_name].append(value)

    @abc.abstractmethod
    async def execute(
        self,
        custom_data: typing.Optional[dict] = None,
        event_type: typing.Optional[str] = None,
    ):
        """
        Commit upsertor data to the storage. Afterwards, send a webhook request with upsertion details.

        Args:
                custom_data: Custom execution data. Included in webhook payload.
                event_type: Event type included in webhook payload.

        Raises:
                DuplicateError: Raised if there is a colliding object already stored in a storage.
        """

        pass

    async def webhook(self, data: dict):
        # TODO: add docstring
        assert self.Storage.WebhookURIs is not None
        json_dump = JSONDumper(pretty=False)(data)
        for uri in self.Storage.WebhookURIs:
            self.WebhookResponseData[uri] = await self.Storage.ProactorService.execute(
                self._webhook, json_dump, uri, self.Storage.WebhookAuth
            )

    def _webhook(self, data, uri, auth=None):
        u = urllib.parse.urlparse(uri)
        if u.scheme == "https":
            conn = http.client.HTTPSConnection(u.netloc)
        else:
            conn = http.client.HTTPConnection(u.netloc)

        headers = {"Content-Type": "application/json"}
        if auth is not None:
            headers["Authorization"] = auth

        try:
            conn.request("PUT", uri, data, headers)
            response = conn.getresponse()
            if response.status // 100 != 2:
                text = response.read()
                L.error(
                    "Webhook endpoint responded with {}: {}".format(
                        response.status, text
                    ),
                    struct_data={"uri": uri},
                )
                return
            self.WebhookResponseData = json.load(response)
        except ConnectionRefusedError:
            L.error(
                "Webhook call failed: Connection refused.", struct_data={"uri": uri}
            )
            return
        except json.decoder.JSONDecodeError as e:
            L.error(
                "Failed to decode JSON response from webhook: {}".format(str(e)),
                struct_data={"uri": uri},
            )
        except Exception as e:
            L.error(
                "Webhook call failed with {}: {}".format(type(e).__name__, str(e)),
                struct_data={"uri": uri},
            )
        finally:
            conn.close()



import logging

from ..abc.processor import Processor

##

L = logging.getLogger(__name__)

##


class TimeDriftFilter(Processor):
    """
    TimeDriftFilter drops events, whose @timestamp (in seconds) attribute is older than current time plus configured max_time_drift (in seconds).
    """

    ConfigDefaults = {
        "max_time_drift": 60,
    }

    def __init__(self, app, pipeline, id=None, config=None):
        super().__init__(app, pipeline, id, config)

        self.App = app
        self.MaxTimeDrift = float(self.Config["max_time_drift"])

        metrics_service = app.get_service("asab.MetricsService")
        self.TimeDriftFilterCounter = metrics_service.create_counter(
            "timedriftfilter.status",
            tags={},
            init_values={
                "event.in": 0,
                "event.out": 0,
                "event.drop": 0,
                "timestamp.error": 0,
            },
        )

    def process(self, context, event):
        self.TimeDriftFilterCounter.add("event.in", 1)

        timestamp = event.get("@timestamp")
        if timestamp is None:
            self.TimeDriftFilterCounter.add("timestamp.error", 1)
            return None

        difference = self.App.time() - timestamp
        if difference > self.MaxTimeDrift:
            self.TimeDriftFilterCounter.add("event.drop", 1)
            return None

        self.TimeDriftFilterCounter.add("event.out", 1)
        return event



import logging

from ...abc import Expression
from ..value.valueexpr import VALUE

###

L = logging.getLogger(__name__)

###


class SPLIT(Expression):
    Attributes = {
        "Value": ["str"],
        "Separator": ["str"],
    }

    Category = "String"

    def __init__(self, app, *, arg_value, arg_separator):
        super().__init__(app)

        if isinstance(arg_value, Expression):
            self.Value = arg_value
        else:
            self.Value = VALUE(app, value=arg_value)

        if isinstance(arg_separator, Expression):
            self.Separator = arg_separator
        else:
            self.Separator = VALUE(app, value=arg_separator)

    def __call__(self, context, event, *args, **kwargs):
        value = self.Value(context, event, *args, **kwargs)
        separator = self.Separator(context, event, *args, **kwargs)

        try:
            return value.split(separator)
        except AttributeError:
            return None



""" Module for connecting to Mongo database"""

import asyncio
import logging
from bspump.abc.source import TriggerSource

#

L = logging.getLogger(__name__)


#


class MongoDBSource(TriggerSource):
    """MongoDB database source"""

    ConfigDefaults = {
        "output_queue_max_size": 2,
        "database": "",
        "collection": "",
    }

    def __init__(
        self, app, pipeline, connection, query_parms=None, id=None, config=None
    ):
        """
        Create new instance
                Parameters:
                        app (Application): application
                        pipeline (PipeLone): pipeline
                        connection (Connection): Connection to the database
                        query_parms (Dictionary): Query parameters (filter,projection,number of records)
                        id (int): Id
                        config (Dictionary): Connection configuration
        """

        super().__init__(app, pipeline, id=id, config=config)

        self.Pipeline = pipeline
        self._output_queue = asyncio.Queue()
        self._output_queue_max_size = int(self.Config["output_queue_max_size"])
        self.QueryParms = query_parms
        self.Connection = pipeline.locate_connection(app, connection)
        self.Database = self.Config["database"]
        self.Collection = self.Config["collection"]

    async def cycle(self):
        db = self.Connection.Client[self.Database]

        # We check the queue size and remove throttling if the size is smaller than its defined max size.
        if self._output_queue.qsize() == self._output_queue_max_size - 1:
            self.Pipeline.throttle(self, False)

        coll = db[self.Collection]
        await self.Pipeline.ready()

        # query parms
        q_filter = self.QueryParms.get("filter", None)
        q_projection = self.QueryParms.get("projection", None)
        q_limit = self.QueryParms.get("limit", 0)

        cur = coll.find(q_filter, q_projection, 0, int(q_limit))

        async for recs in cur:
            pass
            await self.process(recs, context={})



from ...abc import Expression


class VALUE(Expression):
    """
    Returns specified **scalar** value
    """

    Attributes = {}

    Category = "Value"

    def __init__(self, app, *, value, outlet_type=None):
        super().__init__(app)
        assert not isinstance(value, Expression)
        self.Value = value

        if outlet_type is None:
            self.OutletType = type(self.Value).__name__
        else:
            self.OutletType = outlet_type

    def __call__(self, context, event, *args, **kwargs):
        return self.Value

    def get_outlet_type(self):
        return self.OutletType



import asyncio
import datetime
import logging
import random
import re
import typing

from ..abc.sink import Sink

#

L = logging.getLogger(__name__)

#


class SFTPSink(Sink):
    """

            SFTPSink is a sink processor that forwards events to a remote files created on SSH server.

            SFTPSink expects bytes as an input. If the input is string, it is automatically transformed to bytes.

    .. code:: python

            class SamplePipeline(bspump.Pipeline):
                    def __init__(self, app, pipeline_id=None):
                            super().__init__(app, pipeline_id)
                            self.build(
                                    bspump.random.RandomSource(app, self, choice=['ab', 'bc', 'cd', 'de', 'ef', 'fg'], config={'number': 50}
                                                                                    ).on(bspump.trigger.OpportunisticTrigger(app, chilldown_period=60)),
                                    bspump.common.StdDictToJsonParser(app,self),
                                    bspump.ssh.SFTPSink(app, self, "SSHConnection", config={'remote_path': '/test_folder/',
                                                                                                                                                    'filename': 'myFileToUpload',
                                                                                                                                                    'mode': 'a',
                                                                                                                                                    })
                            )

    """

    ConfigDefaults = {
        "remote_path": "/upload/",
        "prefix": "prefix",
        "filename": "hostname",
        "suffix": "suffix",
        "mode": "a",  # w = write, a = append
    }

    def __init__(self, app, pipeline, connection, id=None, config=None):
        super().__init__(app, pipeline, id=id, config=config)

        self.Connection = pipeline.locate_connection(app, connection)

        self.RemotePath = self.Config["remote_path"]
        self.Prefix = self.Config["prefix"]
        self.FileName = self.Config["filename"]
        self.Suffix = self.Config["suffix"]
        self.RandIntLen = 1000
        self.Mode = self.Config["mode"]
        self.Pipeline = pipeline

        metrics_service = app.get_service("asab.MetricsService")
        self.RewriteCounter = metrics_service.create_counter(
            "sink.rewritefile", tags={}, init_values={"times": 0}
        )

        self._output_queue = asyncio.Queue()
        self._output_queue_max_size = 1000
        self._conn_future = None

        # Subscription
        self._on_health_check("connection.open!")

        app.PubSub.subscribe("Application.stop!", self._on_application_stop)
        app.PubSub.subscribe("Application.tick!", self._on_health_check)
        app.PubSub.subscribe("Application.exit!", self._on_exit)

    def _on_health_check(self, message_type):
        if self._conn_future is not None:
            # Connection future exists

            if not self._conn_future.done():
                # Connection future didn't result yet
                # No sanitization needed
                return

            try:
                self._conn_future.result()
            except Exception:
                # Connection future threw an error
                L.exception("Unexpected connection future error")

            # Connection future already resulted (with or without exception)
            self._conn_future = None

        assert self._conn_future is None

        self._conn_future = asyncio.ensure_future(self.outbound(), loop=self.Loop)

    def _on_application_stop(self, message_type, counter):
        self._output_queue.put_nowait((None, None, None))

    async def _on_exit(self, message_type):
        if self._conn_future is not None:
            await asyncio.wait([self._conn_future], return_when=asyncio.ALL_COMPLETED)

    async def outbound(self):
        async with self.Connection.run() as connection:
            while True:
                async with connection.start_sftp_client() as sftp:
                    event, remote, filename = await self._output_queue.get()

                    if event is None and remote is None and filename is None:
                        break

                    if self._output_queue.qsize() == self._output_queue_max_size - 1:
                        self.Pipeline.throttle(self, False)

                    # Checks if the file of given name already exists on remote folder, and if so, prints the message
                    # to a log file
                    if await sftp.exists(remote + filename) and self.Mode == "w":
                        self.RewriteCounter.add("times", 1)
                        L.warning("File {} has been overwritten".format(str(filename)))

                    # Writes event into a remote file
                    async with sftp.open(
                        remote + filename, self.Mode, encoding=None
                    ) as sftpfile:
                        # TODO fix RuntimeError when emptying queue on kill
                        await sftpfile.write(event)

    def process(self, context, event: typing.Union[dict, str, bytes]):
        # Checks bytes in the event
        if type(event) == str:
            event = event.encode("utf-8")
        elif type(event) == bytes:
            event = event

        if self.FileName is None or self.FileName == "":
            self.FileName = self.build_remote_file_name()

        # Gets the defined remote path and filename from context
        remote = context.get("ssh.remote", self.RemotePath)
        filename = context.get("ssh.filename", self.FileName)

        self._output_queue.put_nowait((event, remote, filename))

        if self._output_queue.qsize() == self._output_queue_max_size:
            self.Pipeline.throttle(self, True)

    def build_remote_file_name(self):
        # Create random string around the user defined name core
        basename = re.sub("[/,.,:, ]", "", str(self.FileName))
        timestamp = str(int(datetime.datetime.timestamp(datetime.datetime.now())))
        random_num = str(random.randint(1, self.RandIntLen))
        filename = str(self.Prefix + basename + timestamp + random_num + self.Suffix)

        return filename



import binascii

from cryptography.hazmat.primitives import padding
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.backends import default_backend

from ..abc.processor import Processor


class CommonAESMixin(object):
    ConfigDefaults = {
        "key": "00000000000000000000000000000000",  # Hexadecimal key, It must be 16 (AES-128), 24 (AES-192), or 32 (AES-256) bytes long.
        "iv": "00000000000000000000000000000000",
    }

    def initalize_aes(self):
        backend = default_backend()

        key = binascii.unhexlify(self.Config["key"])
        iv = binascii.unhexlify(self.Config["iv"])

        self.padding = padding.PKCS7(128)
        self.cipher = Cipher(algorithms.AES(key), modes.CBC(iv), backend=backend)


class EncryptAESProcessor(Processor, CommonAESMixin):
    def __init__(self, app, pipeline, id=None, config=None):
        super().__init__(app, pipeline, id, config)
        self.initalize_aes()

    def process(self, context, event):
        encryptor = self.cipher.encryptor()
        padder = self.padding.padder()
        data = padder.update(event) + padder.finalize()
        return encryptor.update(data) + encryptor.finalize()


class DecryptAESProcessor(Processor, CommonAESMixin):
    def __init__(self, app, pipeline, id=None, config=None):
        super().__init__(app, pipeline, id, config)
        self.initalize_aes()

    def process(self, context, event):
        decryptor = self.cipher.decryptor()
        unpadder = self.padding.unpadder()
        data = decryptor.update(event) + decryptor.finalize()
        return unpadder.update(data) + unpadder.finalize()



import os

import bspump.declarative
import bspump.unittest


class TestDeclarativeWhen(bspump.unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.Builder = bspump.declarative.ExpressionBuilder(self.App)

    def load(self, decl_fname):
        basedir = os.path.dirname(__file__)
        with open(os.path.join(basedir, decl_fname), "r") as f:
            return self.Builder.parse(f.read())[0]

    def test_when_01_else(self):
        decl = self.load("./test_when.yaml")
        res = decl({}, {"key": -1})
        self.assertEqual(res, "Unknown")

    def test_when_01_exact(self):
        decl = self.load("./test_when.yaml")
        res = decl({}, {"key": 34})
        self.assertEqual(res, "Thirty four")

    def test_when_01_range(self):
        decl = self.load("./test_when.yaml")
        res = decl({}, {"key": 45})
        self.assertEqual(res, "fourty to fifty (exclusive)")

    def test_when_01_set(self):
        decl = self.load("./test_when.yaml")
        res = decl({}, {"key": 75})
        self.assertEqual(res, "seventy five, seven, nine")



from aiohttp.web import Request
from bspump.http.web.components import BaseField
from bspump.http.web.template_env import env
import json


class TableComponent(BaseField):
    """
    A component that displays data in a table format with configurable columns.
    Supports adding new rows with editable fields and various input types.

    Supported input_type values:
    - "text": Text input for general text
    - "email": Email input with format validation
    - "number": Number input for numeric values
    - "date": Date picker for date selection
    - "datetime": DateTime picker for date and time selection
    - "url": URL input for website links
    - "textarea": Multi-line text area
    - "select": Dropdown selection (requires choices list)
    - "checkbox": Boolean checkbox input

    Example of defining columns:
    cols = [
        {"key": "email", "display": "Email Address", "input_type": "email", "required": True},
        {"key": "first_name", "display": "First Name", "input_type": "text", "required": False},
        {"key": "last_name", "display": "Last Name", "input_type": "text", "required": False},
        {"key": "phone", "display": "Phone Number", "input_type": "text", "required": True},
        {"key": "notes", "display": "Notes", "input_type": "textarea", "required": False},
        {"key": "status", "display": "Status", "input_type": "select", "choices": ["Active", "Inactive", "Pending", "Suspended"], "required": True},
    ]
    """

    def __init__(
        self, name, data_provider=None, columns=None, editable=False, **kwargs
    ):
        super().__init__(name, **kwargs)
        self.field_name = f"f___{self.name}"
        self.data_provider = data_provider
        self.columns = columns or []
        self.editable = editable

    def html(self, default=""):
        data = []
        if self.data_provider:
            try:
                data = self.data_provider()
            except Exception as e:
                data = [{"error": f"Failed to load data: {str(e)}"}]

        template = env.get_template("table.html")
        return template.render(
            name=self.name,
            field_name=self.field_name,
            display=self.display,
            columns=self.columns,
            data=data,
            hidden=self.hidden,
            editable=self.editable,
            default=default,
        )

    def restructure_data(self, dfrom, dto):
        table_data = dfrom.get(self.field_name, "[]")
        try:
            dto[self.name] = json.loads(table_data)
        except (json.JSONDecodeError, TypeError):
            dto[self.name] = []

    def clean(self, data, request: Request = None):
        pass



import typing


class LibraryProviderABC(object):
    def __init__(self, library, layer):
        super().__init__()
        self.App = library.App
        self.Library = library
        self.Layer = layer
        self.IsReady = False

    async def finalize(self, app):
        pass

    async def read(self, path: str) -> typing.IO:
        """
        Reads a library item on the given path.

        :param path: The path to the item to read
        :return: I/O stream (read) with the content of the library item.
        """
        raise NotImplementedError("{}.read()".format(self.__class__.__name__))

    async def list(self, path: str) -> list:
        """
        It lists all items in the library at the given path.

        :param path: The path to the directory in the library to list
        :return: A list (or iterable) of `LibraryItem`s.
        """
        raise NotImplementedError("{}.list()".format(self.__class__.__name__))

    async def _set_ready(self, ready=True):
        self.IsReady = ready
        await self.Library._set_ready(self)

    async def subscribe(self, path: str):
        """
        It takes a path and subscribes to changes in this directory.
        When change occurs, it creates a PubSub signal.
        Use absolute path, startng with "/".
        Keep in mind every provider requires specific implementation. These might vary in lag between change if the librarby and its propagation.

        Mind following when implementing this method:
        - user can subscribe only on existing directory. -> Check whether subscribed path is a directory, specifically in the provider
        - subscribing to nonexisting directory or file should lead to silent error

        """
        raise NotImplementedError("{}.subscribe()".format(self.__class__.__name__))



import base64
import binascii
import json
import logging
import aiohttp
import aiohttp.client_exceptions

import bspump.asab as asab
from bspump.asab import Service, Config
import bspump.asab.exceptions as exceptions

try:
    import jwcrypto.jwk
    import jwcrypto.jwt
    import jwcrypto.jws
except ModuleNotFoundError:
    jwcrypto = None

#

L = logging.getLogger(__name__)

#


Config.add_defaults(
    {
        "authz": {
            "public_keys_url": "",
            "_disable_token_verification": "no",
            "_disable_rbac": "no",
        }
    }
)


class AuthzService(Service):
    """
    OBSOLETE, use 'asab.web.auth' instead
    """

    def __init__(self, app, service_name="asab.AuthzService"):
        asab.LogObsolete.warning(
            "Module 'asab.web.authz' is deprecated, please use 'asab.web.auth' instead.",
            struct_data={"eol": "2024-01-31"},
        )

        super().__init__(app, service_name)
        self.RBACDisabled = Config.getboolean("authz", "_disable_rbac")
        self._TokenVerificationDisabled = Config.getboolean(
            "authz", "_disable_token_verification"
        )
        if jwcrypto is None and not self._TokenVerificationDisabled:
            raise ModuleNotFoundError(
                "You are trying to use asab.web.authz without 'jwcrypto' installed. "
                "Please run 'pip install jwcrypto' "
                "or install asab with 'authz' optional dependency."
            )
        self.PublicKeysUrl = Config.get("authz", "public_keys_url")
        if len(self.PublicKeysUrl) == 0 and not self._TokenVerificationDisabled:
            raise ValueError("No public_keys_url provided in [authz] config section.")
        self.AuthServerPublicKey = None  # TODO: Support multiple public keys
        # TODO: Fetch public keys if validation fails (instead of periodic fetch)
        self.App.PubSub.subscribe(
            "Application.tick/30!", self._fetch_public_keys_if_needed
        )

    async def initialize(self, app):
        await self._fetch_public_keys_if_needed()

    def is_ready(self):
        if self._TokenVerificationDisabled is True:
            return True
        elif self.AuthServerPublicKey is not None:
            return True
        return False

    async def authorize(self, resources, bearer_token, tenant=None):
        # Use userinfo to make RBAC check
        userinfo = self.userinfo(bearer_token)

        # Fail if userinfo cannot be fetched or resources are missing
        if userinfo is None:
            return False
        user_resources = userinfo.get("resources")
        if user_resources is None:
            return False

        # Allow superuser to pass any check
        if "authz:superuser" in frozenset(user_resources.get("*", [])):
            return True

        if tenant is None:
            # Check only global resources if no tenant is specified
            tenant = "*"
        if tenant not in user_resources:
            # Tenant section is not present: The check has failed
            return False

        # Make sure all the required resources are accessible
        tenant_user_resources = frozenset(user_resources[tenant])
        for resource in resources:
            if resource == "tenant:access":
                # Tenant section is present: User has tenant access
                continue
            if resource not in tenant_user_resources:
                return False

        return True

    def userinfo(self, bearer_token):
        if not self.is_ready():
            L.error("AuthzService is not ready: No public keys loaded yet.")
            return None

        if self._TokenVerificationDisabled:
            return _get_id_token_claims_without_verification(bearer_token)
        else:
            return _get_id_token_claims(bearer_token, self.AuthServerPublicKey)

    async def _fetch_public_keys_if_needed(self, *args, **kwargs):
        if self.is_ready():
            return

        async with aiohttp.ClientSession() as session:
            try:
                async with session.get(self.PublicKeysUrl) as response:
                    if response.status != 200:
                        L.error(
                            "HTTP error while loading public keys.",
                            struct_data={
                                "status": response.status,
                                "url": self.PublicKeysUrl,
                                "text": await response.text(),
                            },
                        )
                        return
                    try:
                        data = await response.json()
                    except json.JSONDecodeError:
                        L.error(
                            "JSON decoding error while loading public keys.",
                            struct_data={
                                "url": self.PublicKeysUrl,
                                "data": data,
                            },
                        )
                        return
                    try:
                        key_data = data["keys"].pop()
                    except (IndexError, KeyError):
                        L.error(
                            "Error while loading public keys: No public keys in server response.",
                            struct_data={
                                "url": self.PublicKeysUrl,
                                "data": data,
                            },
                        )
                        return
                    try:
                        public_key = jwcrypto.jwk.JWK(**key_data)
                    except Exception as e:
                        L.error(
                            "JWK decoding error while loading public keys: {}.".format(
                                e
                            ),
                            struct_data={
                                "url": self.PublicKeysUrl,
                                "data": data,
                            },
                        )
                        return
            except aiohttp.client_exceptions.ClientConnectorError as e:
                L.error(
                    "Connection error while loading public keys: {}".format(e),
                    struct_data={
                        "url": self.PublicKeysUrl,
                    },
                )
                return

        self.AuthServerPublicKey = public_key
        L.log(
            asab.LOG_NOTICE,
            "Public key loaded.",
            struct_data={"url": self.PublicKeysUrl},
        )


def _get_id_token_claims(bearer_token: str, auth_server_public_key):
    assert jwcrypto is not None
    try:
        token = jwcrypto.jwt.JWT(jwt=bearer_token, key=auth_server_public_key)
    except jwcrypto.jwt.JWTExpired:
        raise exceptions.NotAuthenticatedError("ID token expired.")
    except jwcrypto.jws.InvalidJWSSignature:
        raise exceptions.NotAuthenticatedError("Invalid ID token signature.")
    except ValueError as e:
        raise exceptions.NotAuthenticatedError("Authentication failed: {}".format(e))

    try:
        token_claims = json.loads(token.claims)
    except ValueError:
        raise exceptions.NotAuthenticatedError("Cannot parse ID token claims.")

    return token_claims


def _get_id_token_claims_without_verification(bearer_token: str):
    try:
        header, payload, signature = bearer_token.split(".")
    except IndexError:
        raise exceptions.NotAuthenticatedError(
            "Cannot parse ID token: Wrong number of '.'."
        )

    try:
        claims = json.loads(base64.b64decode(payload.encode("utf-8")))
    except binascii.Error:
        raise exceptions.NotAuthenticatedError(
            "Cannot parse ID token: Payload is not base 64."
        )
    except json.JSONDecodeError:
        raise exceptions.NotAuthenticatedError(
            "Cannot parse ID token: Payload cannot be parsed as JSON."
        )

    return claims



import re
import json
import logging
import time
from bspump.asab import Service, Config
from bspump.asab.web.rest.json import JSONDumper

L = logging.getLogger(__name__)


def get_pipeline_topology(pump, pipeline):
    dumper = JSONDumper(pretty=False)
    pipelines = json.loads(dumper(pump.Pipelines))

    pipeline_obj = pump.Pipelines[pipeline]

    pipeline_data = pipelines[pipeline]
    components = []
    metrics_components = []
    components.append(pipeline_data["Sources"][0])
    # get metrics
    for metric in pipeline_data["Metrics"]:
        if (
            metric["type"] == "Counter"
            and metric["name"] == "bspump.pipeline.eps_processor"
        ):
            metrics_components.append(metric)
    for depth in pipeline_data["Processors"]:
        for component in depth:
            components.append(component)

    output = {
        "topology": {},
        "display-style": "graph",
        "display-priority": "shown",
    }
    for i in range(len(components)):
        component_data = {}

        component_obj = pipeline_obj.locate_source(components[i]["Id"])
        if component_obj is None:
            component_obj = pipeline_obj.locate_processor(components[i]["Id"])

        # If it's not the last element, add the "wires" key
        if i < len(components) - 1:
            component_data["wires"] = [components[i + 1]["Id"]]
        else:
            component_data["wires"] = []

        # Implement getting properties
        component_data["properties"] = {
            key: value for key, value in component_obj.Config.items()
        }
        component_data["capabilities"] = ["subscribable-events"]

        for metric in metrics_components:
            if metric["static_tags"]["processor"] == components[i]["Id"]:
                component_data["metrics"] = {
                    "eps.in": metric["fieldset"][0]["values"]["eps.in"],
                    "eps.out": metric["fieldset"][0]["values"]["eps.out"],
                }
                break
        output["topology"][components[i]["Id"]] = component_data

    # add metrics to the source as it doesn't have any metrics from the metric service
    if "metrics" not in output["topology"][components[0]["Id"]]:
        if "eps.out" in output["topology"][components[1]["Id"]]["metrics"]:
            output["topology"][components[0]["Id"]]["metrics"] = {}
            output["topology"][components[0]["Id"]]["metrics"]["eps.out"] = int(
                output["topology"][components[1]["Id"]]["metrics"]["eps.out"]
            )

    message = get_message_structure()
    message["data"] = output
    message["count"] = 1
    message["remaining_subscription_count"] = 1

    return message


def get_pipelines(pipelines: dict):
    output = {"topology": {}, "display-style": "graph", "display-priority": "hidden"}
    for pipeline in pipelines.values():
        pipeline_dict = {
            "wires": [],
            "properties": [],
            "metrics": [],
            "capabilities": ["has-children"],
        }
        output["topology"][pipeline.Id] = pipeline_dict

    message = get_message_structure()
    message["data"] = output
    message["count"] = 1
    message["remaining_subscription_count"] = 1

    return message


def get_message_structure():
    return {
        "timestamp": time.time_ns(),
        "data": {},
        "count": 0,
        "remaining_subscription_count": 0,
    }


class MQTTService(Service):
    def __init__(self, app, service_name="bspump.MQTTService", connection=None):
        super().__init__(app, service_name)
        self.App = app
        self.dumper = JSONDumper(pretty=False)
        self.Connection = None
        self.ConnectionId = connection
        try:
            self.max_count = int(Config["mqtt"].get("max_count"))
        except Exception:
            L.warning(
                "MQTTService: The max_count parameter is not set in the configuration file or is not an integer. Default value is 100."
            )
            self.max_count = 100

    def components_initialize(self):
        svc = self.App.get_service("bspump.PumpService")
        self.Connection = svc.locate_connection(self.ConnectionId)

        for pipeline in svc.Pipelines.values():
            self.Connection.publish_to_topic(
                f"/c/{pipeline.Id}/topology",
                json.dumps(get_pipeline_topology(svc, pipeline.Id)),
                retain=True,
            )

            for depth in pipeline.Processors:
                for component in depth:
                    pipeline.PublishingProcessors[component.Id] = 0
                    self.Connection.subscribe_topic(
                        f"/c/{pipeline.Id}/c/{component.Id}/events/subscribe"
                    )
                    self.Connection.register_handler(
                        f"/c/{pipeline.Id}/c/{component.Id}/events/subscribe",
                        self.on_message,
                    )

            for source in pipeline.Sources:
                self.Connection.subscribe_topic(
                    f"/c/{pipeline.Id}/c/{source.Id}/events/subscribe"
                )
                self.Connection.register_handler(
                    f"/c/{pipeline.Id}/c/{source.Id}/events/subscribe",
                    self.on_message,
                )

        self.Connection.publish_to_topic(
            "/topology",
            json.dumps(get_pipelines(svc.Pipelines)),
            retain=True,
        )

    def on_message(self, client, userdata, message):
        payload = message.payload.decode("utf-8")
        svc = self.App.get_service("bspump.PumpService")
        topic = message.topic

        # Regex patterns
        events_pattern = r"^/c/(?P<pipeline_identifier>[^/]+)/c/(?P<component_identifier>[^/]+)/events/subscribe$"

        # Matching
        events = re.match(events_pattern, topic)

        # Get list of pipelines from application
        try:
            payload = json.loads(payload)
        except json.decoder.JSONDecodeError:
            payload = message.payload.decode("utf-8")
            L.warning(
                f"Payload sent to {topic} is not a valid JSON. Payload: {payload}"
            )
            return

        count = payload.get("count")
        if count is None or count < 0:
            return

        count = min(count, self.max_count)

        if events:
            pipeline = events.group("pipeline_identifier")
            processor = events.group("component_identifier")
            pipeline = svc.locate(f"{pipeline}")

            if pipeline is None:
                L.warning(f"Pipeline {pipeline} not found")
                return

            source = pipeline.locate_source(f"{processor}")
            if source is not None:
                source.EventsToPublish = count
            else:
                processor = pipeline.locate_processor(f"{processor}")

                if processor is None:
                    L.warning(f"Processor {processor} not found")
                    return

                pipeline.PublishingProcessors[processor.Id] = max(
                    count, pipeline.PublishingProcessors[processor.Id]
                )

    def publish_event(self, pipeline, component, event, count_remaining):
        data = get_message_structure()
        if isinstance(event, bytes):
            event = event.decode("utf-8")
        data["data"] = event
        data["count"] = component.EventCount
        data["remaining_subscription_count"] = count_remaining
        self.Connection.publish_to_topic(
            f"/c/{pipeline}/c/{component.Id}/events",
            json.dumps(data, default=lambda x: x.__class__.__name__),
        )



import functools
import json
import logging
import uuid
import datetime
import aiohttp.web
import fastjsonschema

from ... import exceptions

#

L = logging.getLogger(__name__)
Lex = logging.getLogger("asab.web")


#

_no_log_headers = {
    "Authorization",
}


def _filter_logging_headers(headers):
    """
    Filter HTTP headers to include in log messages.
    """
    return {k: v for k, v in headers.items() if k not in _no_log_headers}


class JSONDumper(object):
    def __init__(self, pretty):
        self.pretty = pretty

    def __call__(self, obj):
        if self.pretty:
            return json.dumps(obj, indent=4, default=self.default) + "\n"
        else:
            return json.dumps(obj, default=self.default)

    def default(self, o):
        m = getattr(o, "rest_get", None)
        if m is not None:
            return m()

        if isinstance(o, datetime.datetime):
            if o.tzinfo == datetime.timezone.utc:
                # isoformat ends with "+00:00", replace it with "Z"
                return o.isoformat()[:-6] + "Z"
            elif o.tzinfo is not None:
                # The datetime object is timezone-aware but not UTC
                # NOT WANTED, using non-UTC timestamps is not recommended
                return o.isoformat()
            else:
                # The datetime object is timezone-naive -> interpret it as UTC
                return o.isoformat() + "Z"

        elif isinstance(o, bytes):
            return o.hex()

        try:
            return json.JSONEncoder.default(self, o)
        except TypeError:
            return "{}".format(o)


def json_response(request, data, pretty=None, dumps=JSONDumper, **kwargs):
    """
    argument `dumps` allows to specify custom JSON dumper for a serialization.
    The default is `JSONDumper` class.

    ## Pretty Result
    When appending ?pretty to any request made, the JSON returned will be pretty formatted (use it for debugging only!).
    """
    assert issubclass(dumps, JSONDumper)
    pretty = (
        request.query.get("pretty", "no").lower()
        in frozenset(["true", "1", "t", "y", "yes", ""])
        or pretty
    )

    return aiohttp.web.json_response(data, dumps=dumps(pretty), **kwargs)


@aiohttp.web.middleware
async def JsonExceptionMiddleware(request, handler):
    """
    Installation of the handler to a web service:

    websvc.WebApp.middlewares.append(asab.web.rest.JsonExceptionMiddleware)
    """

    try:
        response = await handler(request)
        return response

    # HTTP errors to JSON
    except aiohttp.web.HTTPError as ex:
        if ex.status == 401:
            result = "NOT-AUTHORIZED"
        elif ex.status == 404:
            result = "NOT-FOUND"
        else:
            result = "ERROR"

        respdict = {
            "result": result,
            "message": ex.text,
        }
        if ex.status >= 400:
            euuid = uuid.uuid4()
            respdict["uuid"] = str(euuid)
            struct_data = {
                "uuid": str(euuid),
                "path": request.path,
                "status": ex.status,
                **_filter_logging_headers(request.headers),
            }
            Lex.error(ex, struct_data=struct_data)

        ex.content_type = "application/json"
        ex.text = json.dumps(respdict)
        raise ex

    # KeyError translates to 404
    except KeyError as e:
        euuid = uuid.uuid4()
        struct_data = {
            "uuid": str(euuid),
            "path": request.path,
            "status": 404,
            **_filter_logging_headers(request.headers),
        }
        Lex.warning(
            "KeyError when handling web request", exc_info=True, struct_data=struct_data
        )

        if len(e.args) > 1:
            message = e.args[0].format(*e.args[1:])
        elif len(e.args) == 1 and e.args[0] is not None:
            message = e.args[0]
        else:
            message = "KeyError"

        return json_response(
            request,
            data={
                "result": "NOT-FOUND",
                "message": message,
                "uuid": str(euuid),
            },
            status=404,
        )

    # ValidationError translates to 400
    except exceptions.ValidationError as e:
        euuid = uuid.uuid4()
        struct_data = {
            "uuid": str(euuid),
            "path": request.path,
            "status": 400,
            **_filter_logging_headers(request.headers),
        }
        Lex.warning(
            "ValidationError when handling web request",
            exc_info=True,
            struct_data=struct_data,
        )

        if len(e.args) > 1:
            message = e.args[0].format(*e.args[1:])
        elif len(e.args) == 1 and e.args[0] is not None:
            message = e.args[0]
        else:
            message = "ValidationError"
        return json_response(
            request,
            data={
                "result": "VALIDATION-ERROR",
                "message": message,
                "uuid": str(euuid),
            },
            status=400,
        )

    # Conflict surfaces as HTTP 409 Conflict
    except exceptions.Conflict as e:
        euuid = uuid.uuid4()
        struct_data = {
            "uuid": str(euuid),
            "path": request.path,
            "status": 409,
            **_filter_logging_headers(request.headers),
        }
        Lex.warning(
            "Conflict when handling web request", exc_info=True, struct_data=struct_data
        )

        if len(e.args) > 1:
            message = e.args[0].format(*e.args[1:])
        elif len(e.args) == 1 and e.args[0] is not None:
            message = e.args[0]
        else:
            message = "Conflict"

        return json_response(
            request,
            data={
                "result": "CONFLICT",
                "message": message,
                "uuid": str(euuid),
            },
            status=409,
        )

    # Other errors surface as HTTP 500 Internal server error
    except Exception:
        euuid = uuid.uuid4()
        struct_data = {
            "uuid": str(euuid),
            "path": request.path,
            "status": 500,
            **_filter_logging_headers(request.headers),
        }
        Lex.exception(
            "Exception when handling web request",
            exc_info=True,
            struct_data=struct_data,
        )
        return json_response(
            request,
            data={
                "result": "ERROR",
                "message": "Internal Server Error",
                "uuid": str(euuid),
            },
            status=500,
        )


def json_schema_handler(json_schema, *_args, **_kwargs):
    """
    The json schema handler implements validation of JSON documents by JSON schema.
    The library of fastjsonschema implements JSON schema drafts 04, 06 and 07
    https://horejsek.github.io/python-fastjsonschema/

    Examples of use:

    @asab.web.rest.json_schema_handler({
    'type': 'object',
    'properties': {
            'key1': {'type': 'string'},
            'key2': {'type': 'number'},
    }})
    async def login(self, request, *, json_data):
            ...

    or by specifying json as file

    @asab.web.rest.json_schema_handler('./data/sample_json_schema.json')
    async def login(self, request, *, json_data):
            ...

    Works for `application/json`, `application/x-www-form-urlencoded` and `multipart/form-data` post requests
    """

    def decorator(func):
        # Initializing fastjsonschema.compile method and generating
        # the validation function for validating JSON schema

        # JSON schema set as a dict
        if isinstance(json_schema, dict):
            validate = fastjsonschema.compile(json_schema)

        # JSON schema set in a file
        elif isinstance(json_schema, str):
            with open(json_schema) as f:
                schema = json.load(f)
                validate = fastjsonschema.compile(schema)
        else:
            raise ValueError(
                "JSON schema input must be type <class 'dict'> or type <class 'str'>, "
                "not type {}.".format(type(json_schema))
            )

        form_content_types = frozenset(
            ["", "application/x-www-form-urlencoded", "multipart/form-data"]
        )

        @functools.wraps(func)
        async def validator(*args, **kwargs):
            # Initializing fastjsonschema.compile method and generating
            # the validation function for validating JSON schema
            request = args[-1]
            if request.content_type == "application/json":
                try:
                    data = await request.json()
                except json.decoder.JSONDecodeError:
                    raise aiohttp.web.HTTPBadRequest(
                        reason="Failed to parse JSON request"
                    )
            elif request.content_type in form_content_types:
                multi_dict = await request.post()
                data = {k: v for k, v in multi_dict.items()}
            else:
                raise aiohttp.web.HTTPBadRequest(
                    reason="Unsupported content-type {}".format(request.content_type)
                )
            # Checking the validation on JSON data set
            try:
                validate(data)
                kwargs["json_data"] = data
            except fastjsonschema.exceptions.JsonSchemaException as e:
                raise aiohttp.web.HTTPBadRequest(reason=str(e))
            except Exception as e:
                Lex.error(
                    "JSON validation error. Reason: {}".format(e),
                    struct_data={"path": request.path, "method": request.method},
                )
                raise e

            return await func(*args, **kwargs)

        # This is here for Swagger documentation
        setattr(validator, "json_schema", json_schema)

        return validator

    return decorator



import logging

from ..abc.connection import Connection

#

L = logging.getLogger(__name__)

#


class KafkaConnection(Connection):
    """
            KafkaConnection serves to connect BSPump application with an instance of Apache Kafka messaging system.
            It can later be used by processors to consume or provide user-defined messages.

    .. code:: python

            config = {"bootstrap_servers": "localhost:9092"}
            app = bspump.BSPumpApplication()
            svc = app.get_service("bspump.PumpService")
            svc.add_connection(
                    bspump.kafka.KafkaConnection(app, "KafkaConnection", config)
            )

            Standard Kafka configuration options can be used,
            as specified in librdkafka library,
            where the options are simply passed to:

            https://github.com/edenhill/librdkafka/blob/master/CONFIGURATION.md
    """

    ConfigDefaults = {}

    def __init__(self, app, id=None, config=None):
        super().__init__(app, id=id, config=config)



import logging

#

L = logging.getLogger(__name__)

#


class Storage(object):
    def __init__(self):
        self.Metrics = []

    def add(self, metric_name: str, tags: dict, reset: bool, help: str, unit: str):
        for i in range(len(self.Metrics) - 1, -1, -1):
            metric = self.Metrics[i]
            if metric_name != metric["name"]:
                continue
            if tags != metric["static_tags"]:
                continue

            # There is existing storage for a metrics and we are going to overwrite it.
            # It means that the existing metrics will diapear from output and will be replaced by a new one.
            # It is designed to support e.g. dynamic prebuilds of the classes with the same metrics.
            # In other cases, this path should be avoided by e.g. creation of metrics in init time, not during runtime
            L.debug(
                "The metrics storage '{}/{}' is overriden".format(metric_name, tags)
            )
            del self.Metrics[i]

        metric = dict()
        metric["type"] = None  # Will be filled a bit later
        metric["static_tags"] = tags
        metric["name"] = metric_name
        metric["fieldset"] = list()

        if reset is not None:
            metric["reset"] = reset
        if help is not None:
            metric["help"] = help
        if unit is not None:
            metric["unit"] = unit

        self.Metrics.append(metric)
        return metric

    def clear(self):
        self.Metrics.clear()



class DeclarationError(RuntimeError):
    def __init__(self, original_exception, location=None):
        """
        :param original_exception: message
        :param location: used by expressions
        """

        self.OriginalException = original_exception
        self.Location = location
        super().__init__(original_exception)

    def __str__(self):
        if self.Location is None:
            return super().__str__()
        else:
            return "{}\n{}".format(self.OriginalException, self.Location)



import bspump.unittest
from bspump.integrity import IntegrityEnricher


class TestIntegrityEnricher(bspump.unittest.ProcessorTestCase):
    def test_integrity_enricher_01(self):
        events = [
            (
                None,
                {
                    "number": 124,
                    "string": "foo bar",
                    "bytes": b"bar foo",
                },
            ),
        ]

        self.set_up_processor(IntegrityEnricher)

        output = self.execute(events)

        print(output)

        # self.assertEqual(len(output[0][1]), 32)
        # self.assertEqual(
        # 	output[0][1],
        # 	b'\x86ex\x08\x1a\x1d\xeas\xbeG\xdc\xdeE3K\x17\xb2\xcc\xc0=\x88-\xa2\xd0\xfb\x1eQ\xec\x9d\xdcse'
        # )

    def test_integrity_enricher_02(self):
        events = [
            (
                None,
                {
                    "number": 124,
                    "string": "foo bar",
                    "bytes": b"bar foo",
                    "key\u011b": "value\u011b",
                },
            ),
        ]

        self.set_up_processor(IntegrityEnricher)

        output = self.execute(events)

        print(output)

        # self.assertEqual(len(output[0][1]), 32)
        # self.assertEqual(
        # 	output[0][1],
        # 	b'\x86ex\x08\x1a\x1d\xeas\xbeG\xdc\xdeE3K\x17\xb2\xcc\xc0=\x88-\xa2\xd0\xfb\x1eQ\xec\x9d\xdcse'
        # )



import asyncio
import typing


class Timer(object):
    """
    The relative and optionally repeating timer for asyncio.

    This class is simple relative timer that generate an event after a given time,
    and optionally repeating in regular intervals after that.
    The timer object is initialized as stopped.

    Attributes:
            App (asab.Application): Reference to the ASAB application.
            Handler (asyncio.Coroutine | asyncio.Task): A coroutine or future that will be called when a timer triggers.
            Task (asyncio.Future | asyncio.Task | None): A future that represents the timer task.
            AutoRestart (bool): If `True` then a timer will be automatically restarted after triggering.

    Examples:
    ```python
    class TimerApplication(asab.Application):
            async def initialize(self):
                    self.Timer = asab.Timer(self, self.on_tick, autorestart=True)
                    self.Timer.start(1) #(1)

            async def on_tick(self): #(2)
                    print("Think!")
    ```

    1. The timer will trigger a message publishing at every second.
    2. This function has to be a coroutine.

    !!! note
            The implementation idea was borrowed from [StackOverflow discussion](https://stackoverflow.com/questions/45419723/python-timer-with-asyncio-coroutine).
    """

    def __init__(self, app, handler, autorestart=False):
        self.App = app
        self.Handler = handler
        self.Task = None
        self.AutoRestart = autorestart

        app.PubSub.subscribe("Application.stop!", self._on_stop)

    def start(self, timeout: typing.Union[int, float]):
        """
        Start the `Timer` with new timeout.

        Args:
                timeout (int | float): A timer delay in seconds.

        Raises:
                RuntimeError: If the `Timer` has already started.
        """
        if self.is_started():
            raise RuntimeError("Timer is already started")
        self.Task = asyncio.ensure_future(self._job(timeout))

    def stop(self):
        """
        Stop the timer.
        """
        if self.Task is not None:
            self.Task.cancel()
            self.Task = None

    def restart(self, timeout: typing.Union[int, float]):
        """
        Restart the timer with a new timeout.

        Args:
                timeout (int | float): A new timeout in seconds.
        """

        if self.is_started():
            self.stop()
        self.start(timeout)

    def is_started(self) -> bool:
        """
        Check if the `Timer` has started.

        Returns:
                `True` if the `Timer` has started.
        """
        return self.Task is not None

    async def _job(self, timeout):
        await asyncio.sleep(timeout)
        self.Task = None
        if self.AutoRestart:
            self.start(timeout)
        await self.Handler()

    def _on_stop(self, message_type, n):
        # This is to ensure timer stop on application exit
        self.stop()



import os

import bspump.declarative
import bspump.unittest


class TestDeclarativeNestedExpression(bspump.unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.Builder = bspump.declarative.ExpressionBuilder(self.App)
        self.Optimizer = bspump.declarative.ExpressionOptimizer(self.App)

    def load(self, decl_fname):
        basedir = os.path.dirname(__file__)
        with open(os.path.join(basedir, decl_fname), "r") as f:
            return self.Builder.parse(f.read())[0]

    def test_01(self):
        decl = self.load("./test_declarative_nested_expression.yaml")

        event = {
            "continue": "yes",
        }

        res_predicate = self.Optimizer.optimize(decl["predicate"])({}, event)
        res_parse = decl["parse"]({}, event)

        self.assertEqual(res_predicate, True)

        self.assertEqual(
            res_parse,
            {
                "name": "Parse Dict Success",
            },
        )

    def test_02(self):
        decl = self.load("./test_declarative_nested_expression.yaml")

        event = {
            "continue": "yes",
        }

        context = {"brain": {"idea": "This is beautiful parsing."}}
        res_parse = self.Optimizer.optimize(decl["parse"])(context, event)

        self.assertEqual(
            res_parse,
            {
                "name": "Parse Dict Success",
                "idea": "This is beautiful parsing.",
            },
        )

    def test_03(self):
        """
        Test JOIN expression initialization
        :return:
        """

        decl = self.load("./test_declarative_nested_expression.yaml")

        res_parse = self.Optimizer.optimize(decl["join"])({}, "Prema")

        self.assertEqual(
            res_parse,
            {
                "name": "Prema",
                "message": "Prema is nice!",
            },
        )



from .connection import KafkaConnection
from .source import KafkaSource
from .sink import KafkaSink
from .keyfilter import KafkaKeyFilter
from .topic_initializer import KafkaTopicInitializer

__all__ = [
    "KafkaConnection",
    "KafkaSource",
    "KafkaSink",
    "KafkaKeyFilter",
    "KafkaTopicInitializer",
]



import asyncio
import logging
import socket
import re

from ..abc.sink import Sink

from .stream import Stream

#

L = logging.getLogger(__name__)

#


class StreamClientSink(Sink):
    """
    Description:

    """

    ConfigDefaults = {
        "address": "127.0.0.1 8888",  # IPv4, IPv6 or unix socket path
        "outbound_queue_max_size": 100,  # Maximum size of the output queue before throttling
    }

    def __init__(self, app, pipeline, id=None, config=None):
        """
        Description:

        """
        super().__init__(app, pipeline, id=id, config=config)
        self.OutboundQueue = asyncio.Queue()

        # Throttle till we are connected
        self.Pipeline.throttle(self, enable=True)

        # Maximum size for the queue
        self.OutboundQueueMaxSize = int(self.Config["outbound_queue_max_size"])
        assert self.OutboundQueueMaxSize >= 1

        self.Task = None

        self.Pipeline.PubSub.subscribe("bspump.pipeline.start!", self._open_connection)
        self.Pipeline.PubSub.subscribe("bspump.pipeline.stop!", self._close_connection)
        app.PubSub.subscribe("Application.tick!", self._on_tick)

    async def _open_connection(self, message, pipeline):
        """
        Description:

        """
        # Connection is established
        if self.Task is not None:
            await self._close_connection(message, pipeline)

        self.Task = self.Pipeline.Loop.create_task(self._client_connected_task())

    async def _close_connection(self, message, pipeline):
        """
        Description:

        """
        self.Pipeline.throttle(self, enable=True)
        if self.Task is not None:
            self.Task.cancel()
            self.Task = None

    def _on_tick(self, event_name):
        """
        Description:

        """
        # Unthrottle the queue if needed
        if (
            self.OutboundQueue in self.Pipeline.get_throttles()
            and self.OutboundQueue.qsize() < self.OutboundQueueMaxSize
        ):
            self.Pipeline.throttle(self.OutboundQueue, False)

        if self.Task is not None and self.Task.done():
            # We should be connected but we are not
            # Let's do a bit of clean-up and commence reconnection

            try:
                self.Task.result()
            except Exception:
                L.exception("Error when handling client socket")

            self.Task = self.Pipeline.Loop.create_task(self._client_connected_task())

    async def _client_connected_task(self):
        """
        Description:

        """
        loop = self.Pipeline.Loop

        addr = self.Config["address"]

        if " " in addr:
            addr = re.split(r"\s+", addr)
        else:
            # This line allows the (obsolete) format of IPv4 with ':'
            # such as "0.0.0.0:8001"
            addr = re.split(r"[:\s]", addr, 1)

        host = addr.pop(0).strip()
        port = addr.pop(0).strip()
        port = int(port)

        addrinfo = await loop.getaddrinfo(
            host, port, family=socket.AF_UNSPEC, type=socket.SOCK_STREAM
        )

        client_sock = None
        connection_exception = ""
        for family, socktype, proto, canonname, sockaddr in addrinfo:
            client_sock = socket.socket(family, socktype, proto)
            client_sock.setblocking(False)
            try:
                await loop.sock_connect(client_sock, sockaddr)
            except Exception as e:
                connection_exception = e
                client_sock = None
                continue

        if client_sock is None:
            L.warning(
                "Connection to '{}' failed: {}".format(addr, connection_exception)
            )
            return

        # TODO: Support also TLSStream ...
        stream = Stream(
            self.Pipeline.Loop, client_sock, outbound_queue=self.OutboundQueue
        )

        outbound = loop.create_task(stream.outbound())
        inbound = loop.create_task(self._client_inbound_task(client_sock))

        self.Pipeline.throttle(self, enable=False)
        try:
            done, active = await asyncio.wait(
                {outbound, inbound}, return_when=asyncio.FIRST_COMPLETED
            )

        except asyncio.CancelledError:
            active = {outbound, inbound}
            done = {}

        finally:
            self.Pipeline.throttle(self, enable=True)

        # Cancel remaining active tasks
        for t in active:
            # TODO: There could be outstanding data in the outbound queue
            #       Consider flushing them
            t.cancel()

        # Collect results from completed tasks
        for t in done:
            try:
                await t
            except asyncio.CancelledError:
                pass
            except Exception:
                L.exception("Error when handling client socket")

        # Close the stream
        await stream.close()

    async def _client_inbound_task(self, client_sock):
        """
        Description:

        """
        while True:
            data = await self.Pipeline.Loop.sock_recv(client_sock, 4096)
            if len(data) == 0:
                # Client closed the connection
                return

            # Incoming data are discarted ...

    def process(self, context, event):
        """
        Description:

        """
        self.OutboundQueue.put_nowait(event)

        if self.OutboundQueue.qsize() == self.OutboundQueueMaxSize:
            self.Pipeline.throttle(self.OutboundQueue, True)



import os

import bspump.declarative
import bspump.unittest


class TestDeclarativeJoin(bspump.unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.Builder = bspump.declarative.ExpressionBuilder(self.App)

    def load(self, decl_fname):
        basedir = os.path.dirname(__file__)
        with open(os.path.join(basedir, decl_fname), "r") as f:
            return self.Builder.parse(f.read())[0]

    def test_join_01(self):
        event = {
            "string": "STRING",
            "integer": 15,
        }

        decl = self.load("./test_join_01.yaml")

        res = decl({}, event)
        self.assertEqual(res, "STRING:15:")

    def test_join_02(self):
        event = {
            "string": "STRING",
            "integer": 15,
        }

        decl = self.load("./test_join_02.yaml")

        res = decl({}, event)
        self.assertEqual(res, None)



from .aes import DecryptAESProcessor
from .aes import EncryptAESProcessor
from .hashing import HashingProcessor, CoHashingProcessor

"""
Test AES

openssl aes-128-cbc -e -in /etc/services -K 00000000000000000000000000000000 -iv 00000000000000000000000000000000

"""

__all__ = (
    "EncryptAESProcessor",
    "DecryptAESProcessor",
    "HashingProcessor",
    "CoHashingProcessor",
)



from bspump import Sink
import asyncio
import logging


L = logging.getLogger(__name__)


class MongoDBSink(Sink):
    """
    MongoDBSink is a sink processor that forwards the event to a MongoDB specified
    by a MongoDBConnection object.

    MongoDBSink expects either a dictionary or a list of dictionaries as an input.

    Example code can be found in the examples section under bspump-mongo-sink.py

    While the connection defines MongoDB database used, you need to specify particular collection
    inside of this database in the sink itself by modifying the ConfigDefaults while instantiating
    the class.

    """

    ConfigDefaults = {
        "output_queue_max_size": 100,
        "collection": "collection",  # default collection, if not specified inside context
    }

    def __init__(self, app, pipeline, connection, id=None, config=None):
        super().__init__(app, pipeline, id, config)
        # We make use of a connection and pipeline, defined in a different place of the pump.
        self.Connection = pipeline.locate_connection(app, connection)
        self.Pipeline = pipeline

        self._output_queue = asyncio.Queue()
        self._output_queue_max_size = int(self.Config["output_queue_max_size"])
        self.Collection = self.Config["collection"]
        assert self._output_queue_max_size >= 1, "Output queue max size invalid"
        self._conn_future = None

        self._on_health_check("Connection.open!")
        # This part subscribes to outside events used to control the flow of the whole pump.
        # Depending on the specific event, a related class method gets called.
        app.PubSub.subscribe("Application.stop!", self._on_application_stop)
        app.PubSub.subscribe("Application.tick!", self._on_health_check)
        app.PubSub.subscribe("Application.exit!", self._on_exit)

    def _on_health_check(self, message_type):
        # At this point we examine the state of the _conn_future instance variable, queueing _outfluxing
        # in case it is None, returning early if the future is not done and resetting _conn_future to None otherwise.
        if self._conn_future is not None:
            if not self._conn_future.done():
                return
            try:
                self._conn_future.result()
            except Exception:
                # Connection future threw an error
                L.exception("Unexpected connection future error")

            self._conn_future = None

        assert self._conn_future is None

        self._conn_future = asyncio.ensure_future(
            self._insert(),
        )

    def _on_application_stop(self, message_type, counter):
        # On requested stop, we insert 'None' to the FIFO (first in first out) asyncio Queue this means (see later),
        # that no task or item inserted after will be processed.
        self._output_queue.put_nowait((None, None))

    async def _on_exit(self, message_type):
        # On application exit, we first await completion of all tasks in the queue.
        if self._conn_future is not None:
            await asyncio.wait([self._conn_future], return_when=asyncio.ALL_COMPLETED)

    def process(self, context, event: [dict, list]):
        # This is where we check if the queue is overflowing in which case we apply throttling.
        if self._output_queue.qsize() >= self._output_queue_max_size:
            self.Pipeline.throttle(self, True)
        # Here,  we take the event (in this case it should be either a dictionary or a list of dictionaries,
        # and insert it into the queue to be available to the _outflux method.
        self._output_queue.put_nowait((context, event))

    async def _insert(self):
        db = self.Connection.Client[self.Connection.Database]

        while True:
            # Here is where we await the event (which in this case contains the data we want to save to the database)
            # to be pulled out of the queue and saved to a local variable.
            context, event = await self._output_queue.get()

            if event is None:
                break

            # We check the queue size and remove throttling if the size is smaller than its defined max size.
            if self._output_queue.qsize() == self._output_queue_max_size - 1:
                self.Pipeline.throttle(self, False)

            # Obtain the collection to put the obtained item to
            collection = db[context.get("collection", self.Collection)]

            # At this point, we check what kind of item we pulled from the queue. If its None, it means we
            # reached what was inserted by _on_application_stop and we break the loop immediately.
            # If this is not the case we continue determining the type of the item and proceed accordingly,
            # including raising a type error, if the type is unexpected.

            if type(event) == dict:
                await collection.insert_one(event)
                self._output_queue.task_done()
            elif type(event) == list and len(event) > 0:
                await collection.insert_many(event)
                self._output_queue.task_done()
            else:
                raise TypeError(
                    f"Only dict or list of dicts allowed, {type(event)} supplied"
                )



import logging

from ...abc import Expression
from ..value.valueexpr import VALUE
from ...declerror import DeclarationError

###

L = logging.getLogger(__name__)

###


class JOIN(Expression):
    Attributes = {
        "Char": ["str"],
        "Miss": ["*"],  # TODO: This ...
    }

    Category = "String"

    def __init__(self, app, *, arg_items, arg_delimiter=" ", arg_miss=""):
        super().__init__(app)
        self.App = app

        self.Items = arg_items
        self.ItemsNormalized = []

        self.Char = arg_delimiter

        if not isinstance(arg_miss, Expression):
            self.Miss = VALUE(app, value=arg_miss)
        else:
            self.Miss = arg_miss

    def get_outlet_type(self):
        return str.__name__

    def consult_inlet_type(self, key, child):
        return str.__name__

    def set(self, key, value):
        setattr(self, key, value)

        if "Item" in key:
            self.ItemsNormalized[int(key[4:])] = value

    def initialize(self):
        for n, item in enumerate(self.Items):
            if not isinstance(item, Expression):
                item = VALUE(self.App, value=item)

            attr_name = "Item{}".format(n)
            setattr(self, attr_name, item)
            self.Attributes[attr_name] = str.__name__

            self.ItemsNormalized.append(item)

    def __call__(self, context, event, *args, **kwargs):
        try:
            arr = []
            for item in self.ItemsNormalized:
                v = item(context, event, *args, **kwargs)
                if v is None:
                    v = self.Miss(context, event, *args, **kwargs)
                    if v is None:
                        return None
                arr.append(str(v))

            return self.Char.join(arr)

        except Exception as e:
            L.exception(
                "The following exception ocurred in !JOIN expression [delimiter: {}]".format(
                    self.Char
                )
            )
            raise DeclarationError(original_exception=e, location=self.get_location())



# This is the backend for ASAB Web UI Kit.
import os
import glob
import functools

import aiohttp.web


class StaticDirProvider(object):
    def __init__(self, webapp, root, path, index="index.html"):
        self.Path = path
        self.Index = index
        self.Root = root.rstrip("/")

        webapp.router.add_get(self.Root if len(self.Root) > 0 else "/", self._index)

        # Read all files from path (build from React) and map them statically
        for fname in glob.glob(os.path.join(self.Path, "*")):
            if os.path.isfile(fname):
                webapp.router.add_get(
                    self.Root + "/" + os.path.basename(fname),
                    functools.partial(self._file, fname),
                )
            elif os.path.isdir(fname):
                webapp.router.add_static(
                    self.Root + "/" + os.path.basename(fname), path=fname
                )

    async def _index(self, request):
        return aiohttp.web.FileResponse(os.path.join(self.Path, self.Index))

    async def _file(self, fname, request):
        return aiohttp.web.FileResponse(fname)



import json
import yaml
import logging
import typing

import confluent_kafka.admin

import bspump.asab as asab
import bspump
import bspump.abc.sink
import bspump.abc.source


#

L = logging.getLogger(__name__)

#

_TOPIC_CONFIG_OPTIONS = {
    "compression.type",
    "leader.replication.throttled.replicas",
    "message.downconversion.enable",
    "min.insync.replicas",
    "segment.jitter.ms",
    "cleanup.policy",
    "flush.ms",
    "follower.replication.throttled.replicas",
    "segment.bytes",
    "retention.ms",
    "flush.messages",
    "message.format.version",
    "max.compaction.lag.ms",
    "file.delete.delay.ms",
    "max.message.bytes",
    "min.compaction.lag.ms",
    "message.timestamp.type",
    "preallocate",
    "min.cleanable.dirty.ratio",
    "index.interval.bytes",
    "unclean.leader.election.enable",
    "retention.bytes",
    "delete.retention.ms",
    "segment.ms",
    "message.timestamp.difference.max.ms",
    "segment.index.bytes",
}


def _is_kafka_component(component):
    if isinstance(component, bspump.kafka.KafkaSource) or isinstance(
        component, bspump.kafka.KafkaSink
    ):
        return True
    return False


class KafkaTopicInitializer(asab.Configurable):
    """
    KafkaTopicInitializer reads topic configs from file or from Kafka sink/source configs,
    checks if they exists and creates them if they don't.

    KafkaAdminClient requires blocking connection, which is why this class doesn't use
    the connection module from BSPump.

    Usage:
    topic_initializer = KafkaTopicInitializer(app, "KafkaConnection")
    topic_initializer.include_topics(MyPipeline)
    topic_initializer.initialize_topics()
    """

    ConfigDefaults = {
        "client_id": "bspump-topic-initializer",
        "topics_file": "",
        "num_partitions_default": 1,
        "replication_factor_default": 1,
    }

    def __init__(
        self, app, connection, id: typing.Optional[str] = None, config: dict = None
    ):
        """
        Initializes the parameters passed to the class.

        **Parameters**

        app : Application
                Name of the `Application <https://asab.readthedocs.io/en/latest/asab/application.html#>`_.

        connection : Connection
                Information needed to create a connection.

        id: typing.Optional[str] = None :

        config: dict = None : JSON
                configuration file containing important information.

        """
        _id = id if id is not None else self.__class__.__name__
        super().__init__(_id, config)

        # Keeps the topic objects that need to be checked/initialized
        self.RequiredTopics = {}

        # Caches the names of existing topics
        self.ExistingTopics: typing.Optional[set] = None

        self.BootstrapServers = None
        self.ClientId = self.Config.get("client_id")

        self._get_bootstrap_servers(app, connection)

        topics_file = self.Config.get("topics_file")
        if len(topics_file) != 0:
            self.include_topics_from_file(topics_file)

    def _get_bootstrap_servers(self, app, connection):
        svc = app.get_service("bspump.PumpService")
        self.BootstrapServers = (
            svc.Connections[connection].Config["bootstrap_servers"].strip()
        )

    def include_topics(
        self,
        *,
        topic_config=None,
        kafka_component=None,
        pipeline=None,
        config_file=None
    ):
        """
        Includes topic from config file or dict object. It can also scan Pipeline and get topics from Source or Sink.

        **Parameters**

        * :

        topic_config : , default= None
                Topic config file.

        kafka_component : , default= None

        pipeline : , default= None
                Name of the Pipeline.

        config_file : , default= None
                Configuration file.

        """
        # Include topic from config or dict object
        if topic_config is not None:
            L.info("Including topics from dictionary")
            self.include_topics_from_config(topic_config)

        # Include topic from config file
        if config_file is not None:
            L.info("Including topics from '{}'".format(config_file))
            self.include_topics_from_file(config_file)

        # Get topics from Kafka Source or Sink
        if kafka_component is not None:
            L.info("Including topics from {}".format(kafka_component.Id))
            self.include_topics_from_config(kafka_component.Config)

        # Scan the pipeline for KafkaSource(s) or KafkaSink
        if pipeline is not None:
            for source in pipeline.Sources:
                if _is_kafka_component(source):
                    L.info("Including topics from {}".format(source.Id))
                    self.include_topics_from_config(source.Config)
            sink = pipeline.Processors[0][-1]
            if _is_kafka_component(sink):
                L.info("Including topics from {}".format(sink.Id))
                self.include_topics_from_config(sink.Config)

    def include_topics_from_file(self, topics_file: str):
        """
        Includes topics from a topic file.

        **Parameters**

        topics_file:str : str
                Name of a topic file we wanted to include.

        """
        # Support yaml and json input
        ext = topics_file.strip().split(".")[-1].lower()
        if ext == "json":
            with open(topics_file, "r") as f:
                data = json.load(f)
        elif ext in ("yml", "yaml"):
            with open(topics_file, "r") as f:
                data = yaml.safe_load(f)
        else:
            L.warning("Unsupported extension: '{}'".format(ext))

        for topic in data:
            if "topic" not in topic and "name" in topic:  # BACK-COMPAT
                L.warning("Topic attribute 'name' is deprecated; use 'topic' instead.")
                topic["topic"] = topic.pop("name")
            if "config" not in topic and "topic_configs" in topic:  # BACK-COMPAT
                L.warning(
                    "Topic attribute 'topic_configs' is deprecated; use 'config' instead."
                )
                topic["config"] = topic.pop("topic_configs")
            if "num_partitions" not in topic:
                topic["num_partitions"] = int(self.Config.get("num_partitions_default"))
            if "replication_factor" not in topic:
                topic["replication_factor"] = int(
                    self.Config.get("replication_factor_default")
                )
            self.RequiredTopics[topic["topic"]] = confluent_kafka.admin.NewTopic(
                **topic
            )

    def include_topics_from_config(self, config_object):
        """
        Includes topics using a config

        **Parameters**

        config_object : JSON
                config object containing information about what topics we want to include.

        """
        # Every kafka topic needs to have: name, num_partitions and replication_factor
        topic_names = config_object.get("topic").split(",")

        if "num_partitions" in config_object:
            num_partitions = int(config_object.pop("num_partitions"))
        else:
            num_partitions = int(self.Config.get("num_partitions_default"))

        if "replication_factor" in config_object:
            replication_factor = int(config_object.pop("replication_factor"))
        else:
            replication_factor = int(self.Config.get("replication_factor_default"))

        # Additional configs are optional
        topic_configs = config_object.pop("config", {})
        for config_option in set(config_object.keys()):
            if config_option in _TOPIC_CONFIG_OPTIONS:
                topic_configs[config_option] = config_object.pop(config_option)

        # Create topic objects
        for name in topic_names:
            self.RequiredTopics[name] = confluent_kafka.admin.NewTopic(
                topic=name,
                num_partitions=num_partitions,
                replication_factor=replication_factor,
                config=topic_configs,
            )

    def fetch_existing_topics(self):
        admin_client = confluent_kafka.admin.AdminClient(
            {"bootstrap.servers": self.BootstrapServers, "client.id": self.ClientId}
        )
        result = admin_client.list_topics()
        if result is None:
            raise RuntimeError("Empty response from list_topics")
        self.ExistingTopics = set(result.topics.keys())

    def check_and_initialize(self):
        """
        Initializes new topics and logs a warning.

        """
        L.warning(
            "`check_and_initialize()` is obsoleted, use `initialize_topics()` instead"
        )
        self.initialize_topics()

    def initialize_topics(self):
        """
        Initializes topics  ??

        """
        if len(self.RequiredTopics) == 0:
            L.info("No Kafka topics were required.")
            return

        # Fetch topics if not cached
        if self.ExistingTopics is None:
            try:
                self.fetch_existing_topics()
            except Exception as e:
                L.error("Failed to fetch Kafka topics: {}".format(e))
                return

        # Filter out the topics that already exist
        missing_topics = [
            topic
            for topic in self.RequiredTopics.values()
            if topic.topic not in self.ExistingTopics
        ]

        if len(missing_topics) == 0:
            L.info("No missing Kafka topics to be initialized.")
            return

        try:
            admin_client = confluent_kafka.admin.AdminClient(
                {"bootstrap.servers": self.BootstrapServers, "client.id": self.ClientId}
            )

            # Create topics
            # TODO: update configs of existing topics using `admin_client.alter_configs()`
            tasks = admin_client.create_topics(missing_topics)
            for topic, task in tasks.items():
                task.result()

            # Update existing topic cache
            self.ExistingTopics.update(self.RequiredTopics.keys())

            L.log(
                asab.LOG_NOTICE,
                "Kafka topics created",
                struct_data={
                    "topics": ", ".join(topic.topic for topic in missing_topics)
                },
            )

        except Exception as e:
            L.error("Kafka topic initialization failed: {}".format(e))
